/* This QuadRAM application is a production test that verifies that
 * all RAM locations are unique and can store values.
 *
 * Assumptions:
 *
 *   - QuadRAM plugged in to Arduino Mega 1280 or Mega 2560
 *
 * The behavior is as follows:
 *
 *   - Each one of 16 banks of 32 kbytes each is filled with a pseudorandom sequence of numbers
 *   - The 16 banks are read back to verify the pseudorandom sequence
 *   - On success, the Arduino's LED is turned on steadily, with a brief pulse off every 3 seconds
 *   - On failure, the Arduino's LED blinks rapidly
 *   - The serial port is used for status information (38400 bps)
 *
 * This software is licensed under the GNU General Public License (GPL) Version
 * 3 or later. This license is described at
 * http://www.gnu.org/licenses/gpl.html
 *
 * Application Version 1.0 -- October 2011 Rugged Circuits LLC
 * http://www.ruggedcircuits.com/html/quadram.html
 */

#define NUM_BANKS 16

#include "xmem.h"

int debug = 1;

/*
 * Prefix output with file and line info.
 * Helps to figure out what app is running.
 */

void
identify(const char *theFile, int theLine, const char *msg = 0)
{
    if (debug)
    {
        Serial.print("Mega2560-R3 : ");
        Serial.print(theFile);
        Serial.print(" @ ");
        Serial.print(theLine);
        Serial.print(" : ");
        if (msg)
            Serial.println(msg);
    }
}

char ***bankInfo;
int idx;

/*
 * Throughout this program, the term "low memory", "low RAM" refer to the
 * 2560's 8KB of on-board memory, the terms "high memory", "expansion RAM",
 * "expansion memory", "bank" to the expansion card's RAM.
 * 
 * Test scheme works this way:
 * 
 * Each bank is written to by means of an array of C-type (char *) strings,
 * i.e. an array of char*, represented as char**, not *char[], because
 * the array of char* itself is dynamically allocated. Each char* string's
 * payload resides in expansion memory, and so does the char** array of char *,
 * because memory lineups may differ in each bank.
 * 
 * Other way round, each bank contains a dynamically allocated char** variable,
 * with each individual char* within the char** also dynamically allocated.
 * 
 * In theory, all these char** arrays might reside at different addresses,
 * depending on what operations were applied to the heap before creating
 * each bank's char**. In this particular case that's not the case, because
 * the char** is the first object to be allocated.
 * 
 * To track the location of each bank's char** object, an array of 8 char**
 * pointers (one pointer per bank) is dynamically allocated, referenced by
 * the char*** variable bankInfo, in low memory, not within the expansion,
 * because it needs to be accessible regardless of bank selection.
 * 
 * This is accomplished by allocating bankInfo first, then configuring
 * the 2560 for external memory, with heap relocation into the expansion, so
 * that subsequent heap operations take place in high (expansion) memory.
 */

void setup(void)
{
    int i;
    char msg[1024];
    
    // Open up a serial channel
    Serial.begin(9600);
    Serial.println("\n\n\n\n");
    
    /*
     * set up bankInfo. Xmem isn't initialized just yet, so this
     * allocation goes against low memory.
     */
    
    bankInfo = (char ***)malloc(sizeof(char **) * 8);
    
    sprintf(msg,"idx @ %04x, bankInfo @ %04x --> %d",
            (unsigned int)&idx, (unsigned int)&bankInfo,(unsigned int)bankInfo);
    
    identify(__FILE__,__LINE__,msg);
    
    /*
     * Set up QuadRAM. Passing "true" reconfigures the heap to
     * reside in high memory. One side effect of this is that the
     * stack stays in low memory, i.e. the stack sits beneath
     * the heap, and not, like pretty much everywhere else,
     * above.
     */
    xmem::begin(true);
    
    /*
     * run through the banks. Allocate an array of strings in
     * each, then allocate each string
     */
    
    for (i=0;i<8;i++)
    {
        xmem::setMemoryBank(i);
        bankInfo[i] = (char **)malloc(32*sizeof(char *));
        sprintf(msg,"bankInfo[%d] @ %04x",i,(unsigned int)bankInfo[i]);
        identify(__FILE__,__LINE__,msg);
    }
    
    // Enable on-board LED. We like Blinkenlights.
    pinMode(13, OUTPUT);
    
    identify(__FILE__,__LINE__,"Memory test begins ...");
}

// Blink the Arduino LED quickly to indicate a memory test failure
void blinkfast(void)
{
    digitalWrite(13, HIGH);
    delay(100);
    digitalWrite(13, LOW);
    delay(100);
}


/*
#define STARTADDR ((uint8_t *)0x8000)
#define COUNT (0x8000U-1)
 */

#define STARTADDR ((uint8_t *)0x2800)
#define COUNT (0xd800U-1)

void loop(void)
{
    uint8_t bank;
    int i,b;
    char msg[512], testMsg[512];
    char **lines;
    
    identify(__FILE__,__LINE__,"allocating memory...");    
    
    /*
     * in each bank, allocate 32 strings of up to 1KB in length.
     * Set each string to a unique value, containing at least
     * bank and index of string.
     */
    
    for (b=0;b<8;b++)
    {
        xmem::setMemoryBank(b);
        lines = (char **)bankInfo[b];
        for (i=0;i<32;i++)
        {
            lines[i] = (char *)malloc(1024);
            if (lines[i])
                sprintf(lines[i],"This is line[%d] at %x in bank %d\n",
                        i,(unsigned int)(lines[i]),bank);
            
            sprintf(msg,"Bank %d, lines[%d] @ %04x",b,i,(unsigned int)lines[i]);
            /* identify(__FILE__,__LINE__,msg); */
        }
    }
    
    identify(__FILE__,__LINE__,"reading memory...");
    
    /*
     * compare each string in each bank with the value written
     * in the previous loop. Indicate any differences.
     */
    
    for (b=0;b<8;b++)
    {
        xmem::setMemoryBank(b);
        lines = (char **)bankInfo[b];
        for (i=0;i<32;i++)
        {
            if (lines[i])
            {
                sprintf(testMsg,"This is line[%d] at %x in bank %d\n",
                        i,(unsigned int)(lines[i]),bank);
                if (strcmp(lines[i],testMsg))
                {
                    sprintf(msg,"Mismatch: Bank %d, lines[%d] @ %04x, expected %s, got %s",
                            b,i,(unsigned int)lines[i],testMsg,lines[i]);
                    identify(__FILE__,__LINE__,msg);
                }
            }
        }
    }
    
    identify(__FILE__,__LINE__,"releasing memory...");
    
    /*
     * free each string in each bank. Keep the array of char *
     */
    
    for (b=0;b<8;b++)
    {
        xmem::setMemoryBank(b);
        lines = (char **)bankInfo[b];
        for (i=0;i<32;i++)
            if (lines[i])
                free((void*)lines[i]);
    }
    
    blinkfast();  
    
#if 0
    // Keep the LED on mostly, and pulse off briefly every 3 seconds
    while (1) {
        digitalWrite(13, HIGH);
        delay(3000);
        digitalWrite(13, LOW);
        delay(500);
    }
#endif
}
// vim: syntax=cpp ai ts=2 sw=2 cindent expandtab
